from typing import Set, Optional, TYPE_CHECKING

from ..bancho.constants import ClientPackets, CountryCodes, LevelGraph, Privileges, Mode
from ..bancho.streams import StreamOut

from .status import Status

if TYPE_CHECKING:
    from ..game import Game

import logging


class Player:
    def __init__(self, id: int, name: str = "", game: Optional["Game"] = None) -> None:
        self.id = id
        self.name = name
        self.game = game

        self.timezone = 0
        self.country_code = 0
        self.longitude = 0.0
        self.latitude = 0.0

        self.status = Status()
        self.rscore = 0
        self.acc = 100.0
        self.playcount = 0
        self.tscore = 0
        self.rank = 0
        self.pp = 0

        self.privileges: Privileges = Privileges.Normal
        self.spectators: Set[Player] = set()

        self.cant_spectate = False
        self.silenced = False
        self.dms_blocked = False

        self.last_status = Status()

        self.logger = logging.getLogger(self.name)
        self.logger.disabled = game.logger.disabled

    def __repr__(self) -> str:
        return f'<Player "{self.name}" ({self.id})>'

    def __hash__(self) -> int:
        return self.id

    def __eq__(self, other: object) -> bool:
        return self.id == other.id

    @property
    def mode(self) -> Mode:
        return self.status.mode

    @mode.setter
    def mode(self, value: Mode):
        self.status.mode = value

    @property
    def loaded(self) -> bool:
        return bool(self.name)

    @property
    def country(self) -> str:
        return list(CountryCodes.keys())[self.country_code]

    @property
    def level(self) -> int:
        if self.tscore <= 0:
            return 1

        if self.tscore >= LevelGraph[-1]:
            return 100 + int((self.tscore - LevelGraph[99]) / 100000000000)

        for idx, v in enumerate(LevelGraph, start=0):
            if v > self.tscore:
                return idx

        return 1

    def request_presence(self) -> None:
        """Request a presence update for this player"""
        self.game.bancho.request_presence([self.id])

    def request_stats(self) -> None:
        """Request a stats update for this player"""
        self.game.bancho.request_stats([self.id])

    def avatar(self) -> Optional[bytes]:
        """Get the avatar for this player"""
        return self.game.api.get_avatar(self.id)

    def send_message(self, message: str) -> None:
        """Send a PM to this player"""
        if not (player := self.game.bancho.player):
            return

        if not self.loaded:
            # Presence missing
            self.request_presence()

        stream = StreamOut()
        stream.string(player.name)
        stream.string(message)
        stream.string(self.name)
        stream.s32(player.id)

        if not self.game.disable_chat:
            self.logger.info(
                f'<{player.name}{f" ({player.id})" if player.id else ""}> [{self.name}]: "{message}"'
            )

        self.game.bancho.enqueue(ClientPackets.SEND_PRIVATE_MESSAGE, stream.get())

    def add_friend(self) -> None:
        """Add this player to the friends list"""
        if self.id in self.game.bancho.friends:
            self.game.logger.warning(
                f"Tried to add friend, but you are already friends with {self.name}."
            )
            return

        self.game.logger.info(f"You are now friends with {self.name}.")

        self.game.bancho.friends.add(self.id)
        self.game.bancho.enqueue(
            ClientPackets.FRIEND_ADD, int(self.id).to_bytes(4, "little")
        )

    def remove_friend(self) -> None:
        """Remove this player from the friends list"""
        if self.id not in self.game.bancho.friends:
            self.game.logger.warning(
                f"Tried to remove friend, but you are not friends with {self.name}."
            )
            return

        self.game.logger.info(f"You are no longer friends with {self.name}.")

        self.game.bancho.friends.remove(self.id)
        self.game.bancho.enqueue(
            ClientPackets.FRIEND_REMOVE, int(self.id).to_bytes(4, "little")
        )



from typing import Optional, Callable, Dict, List
from datetime import datetime
from copy import copy

from .bancho import TcpBanchoClient, Packets
from ..events import EventHandler
from ..tasks import TaskManager
from ..api import WebAPI

from ..bancho.constants import ServerPackets
from ..objects.client import ClientInfo
from ..utils import deprecated
from ..tasks import Task

import requests
import hashlib
import logging


class TcpGame:
    """osu! game client (tcp)
    --------------------
    This class is a modified version of the "Game" class, that uses the TCP protocol instead of the HTTP protocol.
    Please don't use this class unless you know what you're doing.
    """

    def __init__(
        self,
        username: str,
        password: str,
        server: str,
        version: int,
        executable_hash: str,
        bancho_ip: str,
        bancho_port: int = 13381,
        tournament: bool = False,
        events: Optional[Dict[ServerPackets, List[Callable]]] = {},
        tasks: Optional[List[Task]] = [],
        disable_chat_logging: bool = False,
        disable_logging: bool = False,
    ) -> None:
        """Parameters
        -------------

        `username`: str
            Your bancho username

        `password`: str
            Your bancho password

        `server`: str
            The server to connect to (e.g. `ppy.sh`)

        `version`: int/float
            Set a custom version
            (b20130815 is the latest known version to work with this protocol)

        `executable_hash`: str
            The hash of your osu!.exe file

        `bancho_ip`: str
            The ip of the bancho server

        `bancho_port`: int
            The port of the bancho server (default: 13381)

        `tournament`: bool
            This allows for multiple clients at the same time (supporter only)

        `events`: dict, optional
            Allows to pass in events, instead of registering them normally

        `tasks`: list, optional
            Allows to pass in pre-defined tasks, just like the `events` parameter

        `disable_logging`: bool
            Disable the logging
        """

        self.username = username
        self.password = password
        self.server = server
        self.version = version
        self.tourney = tournament
        self.disable_chat = disable_chat_logging
        self.version_number = version

        self.logger = logging.getLogger("osu!")
        self.logger.disabled = disable_logging

        self.resolve_version()
        self.session = requests.Session()
        self.session.headers = {"User-Agent": "osu!", "osu-version": self.version}
        self.logger.name = f"osu!-{self.version}"

        self.packets = copy(Packets)
        self.events = EventHandler()
        self.bancho = TcpBanchoClient(self, bancho_ip, bancho_port)
        self.tasks = TaskManager(self)
        self.api = WebAPI(self)

        if events:
            self.events.handlers = events

        if tasks:
            self.tasks.tasks = tasks

        self.client = ClientInfo(self.version, executable_hash)

    def __repr__(self) -> str:
        return f"<osu! {self.version}>"

    @property
    def password_hash(self) -> str:
        """Return md5 hashed password"""
        return hashlib.md5(self.password.encode()).hexdigest()

    @property
    def time(self) -> int:
        """Return current time as ticks"""
        return int((datetime.now() - datetime(1, 1, 1)).total_seconds() * 10000000)

    def run(self, retry=False, exit_on_interrupt=False) -> None:
        if retry:
            # Reinitialize game
            self.__init__(
                self.username,
                self.password,
                self.server,
                self.version_number,
                self.client.hash,
                self.bancho.ip,
                self.bancho.port,
                self.tourney,
                self.events.handlers,
                self.tasks.tasks,
                self.logger.disabled,
            )

        try:
            self.bancho.run()
        except KeyboardInterrupt:
            pass
        finally:
            self.logger.warning("Exiting...")
            self.bancho.exit()
            self.logger.warning("Stopping tasks...")
            self.events.executor.shutdown()
            self.tasks.executor.shutdown()

        if exit_on_interrupt:
            exit(0)

    @deprecated
    async def run_async(self) -> None:
        """Run the game in an async context"""
        import asyncio

        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, self.run, False, False)

    def resolve_version(self) -> None:
        """Ensure a correct client version was set"""
        if type(self.version) not in (float, int):
            raise ValueError("Invalid version number")

        self.version = f"b{self.version}"

        if self.tourney:
            self.version = f"{self.version}tourney"



from typing import Optional, List, TYPE_CHECKING
from dataclasses import dataclass
from datetime import datetime

from ..api.constants import SubmissionStatus, Mods, Mode

if TYPE_CHECKING:
    from ..game import Game


@dataclass
class Score:
    id: int
    username: str
    total_score: int
    max_combo: int
    c50: int
    c100: int
    c300: int
    cMiss: int
    cKatu: int
    cGeki: int
    perfect: bool
    mods: Mods
    user_id: int
    rank: Optional[int]
    date: Optional[datetime]
    mode: Mode
    has_replay: bool

    @classmethod
    def from_string(cls, string: str, mode: Mode) -> "Score":
        lines = string.split("|")

        return Score(
            int(lines[0]),
            lines[1],
            int(lines[2]),
            int(lines[3]),
            int(lines[4]),
            int(lines[5]),
            int(lines[6]),
            int(lines[7]),
            int(lines[8]),
            int(lines[9]),
            lines[10] == "1",
            Mods(int(lines[11])),
            int(lines[12]),
            int(lines[13]) if len(lines[13]) > 0 else 0,
            datetime.fromtimestamp(int(lines[14])),
            mode,
            lines[15] == "1",
        )

    def get_comments(self, game: "Game"):
        return game.api.get_comments(replay_id=self.id, mode=self.mode)

    def get_replay(self, game: "Game"):
        return game.api.get_replay(replay_id=self.id, mode=self.mode)


@dataclass
class ScoreResponse:
    status: SubmissionStatus
    beatmap_id: Optional[int]
    set_id: Optional[int]
    total_scores: int
    global_offset: int
    beatmap_format: Optional[str]
    rating: float
    personal_best: Optional[Score]
    scores: List[Score]
    has_osz2: bool = False
    featured_artist_track_id: Optional[int] = None
    featured_artist_license_text: Optional[str] = None

    @classmethod
    def from_string(cls, string: str, mode: Mode) -> "ScoreResponse":
        result = string.split("\n")

        if len(result) <= 0:
            return

        status_results = result[0].split("|")
        beatmap_status = {
            "-1": SubmissionStatus.NotSubmitted,
            "0": SubmissionStatus.Pending,
            "1": SubmissionStatus.Unknown,
            "2": SubmissionStatus.Ranked,
            "3": SubmissionStatus.Approved,
            "4": SubmissionStatus.Qualified,
            "5": SubmissionStatus.Loved,
        }[status_results[0]]

        beatmap_id = None
        beatmapset_id = None
        total_scores = 0
        featured_artist_track_id = None
        featured_artist_license_text = None
        has_osz2 = status_results[1] == "true"

        if len(status_results) > 2:
            beatmap_id = int(status_results[2])
            beatmapset_id = int(status_results[3])
            total_scores = int(status_results[4] or "0")

        if len(status_results) > 5:
            featured_artist_track_id = int(status_results[5] or "0") or None
            featured_artist_license_text = status_results[6] or None

        offset = 0
        beatmap_format = None
        rating = 0.0

        if len(result) > 1:
            offset = int(result[1])
            beatmap_format = result[2]
            rating = float(result[3])

        personal_best = None
        scores = []

        if len(result) > 4:
            if len(result[4]) > 0:
                personal_best = Score.from_string(result[4], mode)

            for score in result[5:]:
                if not score:
                    continue

                scores.append(Score.from_string(score, mode))

        return ScoreResponse(
            beatmap_status,
            beatmap_id,
            beatmapset_id,
            total_scores,
            offset,
            beatmap_format,
            rating,
            personal_best,
            scores,
            has_osz2,
            featured_artist_track_id,
            featured_artist_license_text,
        )



from typing import List, Dict, Callable, TYPE_CHECKING
from copy import copy

from ..objects.replays import ScoreFrame, ReplayFrame
from ..objects.beatmap import BeatmapInfo
from ..objects.channel import Channel
from ..objects.player import Player

if TYPE_CHECKING:
    from ..game import Game

from .streams import StreamIn
from .constants import (
    ServerPackets,
    ClientPackets,
    ReplayAction,
    StatusAction,
    LoginError,
    Privileges,
    Mode,
    Mods,
)

import threading
import zlib
import time


class PacketHandler:
    def __init__(self) -> None:
        self.handlers: Dict[ServerPackets, List[Callable]] = {}

    def register(self, packet: ServerPackets):
        def wrapper(f: Callable):
            if packet in self.handlers:
                self.handlers[packet].append(f)
            else:
                self.handlers[packet] = [f]
            return f

        return wrapper

    def data_received(self, data: bytes, game: "Game"):
        stream = StreamIn(data)

        while not stream.eof():
            packet = ServerPackets(stream.u16())
            compression = stream.bool()  # unused
            data = stream.read(stream.u32())

            if compression:
                # Compression was used in very early versions of bancho
                data = zlib.decompress(data)

            game.logger.debug(f'Received packet {packet.name} -> "{data}"')

            # Handling packet
            self.packet_received(packet, StreamIn(data), game)

            # Reset stream
            stream = StreamIn(stream.readall())

    def packet_received(self, packet: ServerPackets, data: StreamIn, game: "Game"):
        if packet not in self.handlers:
            game.logger.warning(f'No handler found for "{packet.name}"')
            return

        for handler in self.handlers[packet]:
            try:
                handler(data, game)
            except Exception as exc:
                game.logger.error(
                    f"Error while executing {handler}: {exc}", exc_info=exc
                )


Packets = PacketHandler()


@Packets.register(ServerPackets.PONG)
def tcp_ping(stream: StreamIn, game: "Game"):
    game.bancho.enqueue(ClientPackets.PING)


@Packets.register(ServerPackets.PROTOCOL_VERSION)
def protocol_version(stream: StreamIn, game: "Game"):
    game.bancho.protocol = stream.s32()
    game.events.call(ServerPackets.PROTOCOL_VERSION, game.bancho.protocol)


@Packets.register(ServerPackets.RESTART)
def server_restart(stream: StreamIn, game: "Game"):
    timeout = stream.s32() / 1000

    game.logger.warning(f"Bancho is restarting. Retrying in {timeout} seconds...")
    game.bancho.connected = False
    game.bancho.retry = True

    game.events.call(ServerPackets.RESTART, game.bancho.protocol)
    time.sleep(timeout)


@Packets.register(ServerPackets.USER_ID)
def login_reply(stream: StreamIn, game: "Game"):
    response = stream.s32()

    # Check for login errors
    if response < 0:
        try:
            error = LoginError(response)
        except ValueError:
            error = None

        game.logger.error(f"Login error: {error.name}")
        game.logger.error(error.description)
        game.bancho.connected = False
        game.bancho.retry = False

        if error == LoginError.SERVER_ERROR:
            game.bancho.retry = True

        elif error == LoginError.VERIFICATION_NEEDED:
            game.api.verify(game.client.hash)

        return

    game.logger.info(f"Logged in with id: {response}")
    game.bancho.user_id = response

    game.bancho.player = Player(response, game.username, game)
    game.bancho.players.add(game.bancho.player)

    game.bancho.fast_read = True
    game.events.call(ServerPackets.USER_ID, response)


@Packets.register(ServerPackets.PONG)
def pong(stream: StreamIn, game: "Game"):
    game.events.call(ServerPackets.PONG)


@Packets.register(ServerPackets.ACCOUNT_RESTRICTED)
def restricted(stream: StreamIn, game: "Game"):
    game.logger.error("You have been banned.")
    game.events.call(ServerPackets.ACCOUNT_RESTRICTED)
    game.bancho.connected = False
    game.bancho.retry = False


@Packets.register(ServerPackets.PRIVILEGES)
def privileges(stream: StreamIn, game: "Game"):
    game.bancho.privileges = Privileges(stream.s32())
    game.events.call(ServerPackets.PRIVILEGES, game.bancho.privileges)


@Packets.register(ServerPackets.FRIENDS_LIST)
def friends(stream: StreamIn, game: "Game"):
    game.bancho.friends = stream.intlist()
    game.events.call(ServerPackets.FRIENDS_LIST, game.bancho.friends)


@Packets.register(ServerPackets.MAIN_MENU_ICON)
def menu_icon(stream: StreamIn, game: "Game"):
    image, link = stream.string().split("|")
    game.events.call(ServerPackets.MAIN_MENU_ICON, image, link)


@Packets.register(ServerPackets.VERSION_UPDATE)
def version_update(stream: StreamIn, game: "Game"):
    game.logger.info("Bancho requested a version update.")
    game.events.call(ServerPackets.VERSION_UPDATE)


@Packets.register(ServerPackets.VERSION_UPDATE_FORCED)
def version_update_forced(stream: StreamIn, game: "Game"):
    game.logger.warning("Bancho forced a version update.")
    game.events.call(ServerPackets.VERSION_UPDATE_FORCED)


@Packets.register(ServerPackets.GET_ATTENTION)
def attension(stream: StreamIn, game: "Game"):
    game.events.call(ServerPackets.GET_ATTENTION)


@Packets.register(ServerPackets.NOTIFICATION)
def notification(stream: StreamIn, game: "Game"):
    message = stream.string()
    game.logger.info(f'Notification from bancho: "{message}"')
    game.events.call(ServerPackets.NOTIFICATION, message)


@Packets.register(ServerPackets.USER_PRESENCE)
def presence(stream: StreamIn, game: "Game"):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        # Add new player, if not found in collection
        game.bancho.players.add(player := Player(user_id, game=game))

    player.name = stream.string()
    player.timezone = stream.u8() - 24
    player.country_code = stream.u8()

    b = stream.u8()  # Contains privileges and play mode

    player.privileges = Privileges(b & -255)
    player.status.mode = Mode(max(0, min(3, (b & 224) >> 5)))

    player.longitude = stream.float()
    player.latitude = stream.float()
    player.rank = stream.s32()

    game.bancho.fast_read = True
    game.events.call(ServerPackets.USER_PRESENCE, player)


@Packets.register(ServerPackets.USER_STATS)
def stats(stream: StreamIn, game: "Game"):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        # Add new player, if not found in collection
        game.bancho.request_presence([user_id])
        game.bancho.players.add(player := Player(user_id, game=game))

    player.last_status = copy(player.status)

    # Status
    player.status.action = StatusAction(stream.u8())
    player.status.text = stream.string()
    player.status.checksum = stream.string()
    player.status.mods = Mods(stream.u32())
    player.status.mode = Mode(max(0, min(3, stream.u8())))
    player.status.beatmap_id = stream.s32()

    # Stats
    player.rscore = stream.s64()
    player.acc = stream.float()
    player.playcount = stream.s32()
    player.tscore = stream.s64()
    player.rank = stream.s32()
    player.pp = stream.s16()

    game.bancho.fast_read = True
    game.events.call(ServerPackets.USER_STATS, player)


@Packets.register(ServerPackets.USER_PRESENCE_BUNDLE)
def presence_bundle(stream: StreamIn, game: "Game"):
    user_ids = stream.intlist()

    for id in user_ids:
        game.bancho.players.add(Player(id, game=game))

    game.bancho.fast_read = True
    game.events.call(ServerPackets.USER_PRESENCE_BUNDLE, user_ids)


@Packets.register(ServerPackets.USER_PRESENCE_SINGLE)
def presence_single(stream: StreamIn, game: "Game"):
    user_id = stream.s32()

    if not (game.bancho.players.by_id(user_id)):
        # Add player if not found
        game.bancho.players.add(Player(user_id, game=game))

    game.events.call(ServerPackets.USER_PRESENCE_SINGLE, user_id)


@Packets.register(ServerPackets.USER_LOGOUT)
def logout(stream: StreamIn, game: "Game"):
    if not (player := game.bancho.players.by_id(stream.s32())):
        return

    if player is game.bancho.spectating:
        game.logger.info(f"Stopped spectating {player.name}.")
        game.bancho.spectating = None

    game.bancho.players.remove(player)
    game.events.call(ServerPackets.USER_LOGOUT, player)


@Packets.register(ServerPackets.SEND_MESSAGE)
def message(stream: StreamIn, game: "Game"):
    sender = stream.string()
    message = stream.string()
    target = stream.string()
    sender_id = stream.s32()

    # Find player
    if not (player := game.bancho.players.by_id(sender_id)):
        # Try to find sender by name
        if not (player := game.bancho.players.by_name(sender)):
            return

    if not player.loaded:
        # Presence missing
        player.request_presence()

    sender = player

    # Find target
    if target.startswith("#"):
        # Public message
        if not (channel := game.bancho.channels.get(target)):
            return

        target = channel
    else:
        # Private message
        if not (player := game.bancho.players.by_name(target)):
            return

        if not player.loaded:
            # Presence missing
            player.request_presence()

        target = player

    if not game.disable_chat:
        target.logger.info(
            f'<{sender.name}{f" ({sender_id})" if sender_id else ""}> [{target.name}]: "{message}"'
        )

    game.bancho.fast_read = True
    game.events.call(ServerPackets.SEND_MESSAGE, sender, message, target)


@Packets.register(ServerPackets.SPECTATOR_JOINED)
def spectator_joined(stream: StreamIn, game: "Game"):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    player.cant_spectate = False

    game.bancho.player.spectators.add(player)
    game.events.call(ServerPackets.SPECTATOR_JOINED, player)


@Packets.register(ServerPackets.SPECTATOR_LEFT)
def spectator_left(stream: StreamIn, game: "Game"):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    if player not in game.bancho.player.spectators:
        return

    game.bancho.player.spectators.remove(player)
    game.events.call(ServerPackets.SPECTATOR_LEFT, player)


@Packets.register(ServerPackets.FELLOW_SPECTATOR_JOINED)
def fellow_spectator_joined(stream: StreamIn, game: "Game"):
    if not game.bancho.spectating:
        return

    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    game.bancho.spectating.spectators.add(player)
    game.events.call(ServerPackets.FELLOW_SPECTATOR_JOINED, player)


@Packets.register(ServerPackets.FELLOW_SPECTATOR_LEFT)
def fellow_spectator_left(stream: StreamIn, game: "Game"):
    if not game.bancho.spectating:
        return

    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    if player not in game.bancho.spectating.spectators:
        return

    game.bancho.spectating.spectators.remove(player)
    game.events.call(ServerPackets.SPECTATOR_LEFT, player)


@Packets.register(ServerPackets.SPECTATE_FRAMES)
def frames(stream: StreamIn, game: "Game"):
    if not game.bancho.spectating:
        return

    extra = stream.s32()
    frames = [ReplayFrame.decode(stream) for _ in range(stream.u16())]
    action = ReplayAction(stream.u8())

    try:
        score_frame = ScoreFrame.decode(stream)
    except OverflowError:
        score_frame = None

    game.events.call(ServerPackets.SPECTATE_FRAMES, action, frames, score_frame, extra)


@Packets.register(ServerPackets.SPECTATOR_CANT_SPECTATE)
def cant_spectate(stream: StreamIn, game: "Game"):
    if not game.bancho.spectating:
        return

    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    player.cant_spectate = True

    game.events.call(ServerPackets.SPECTATOR_CANT_SPECTATE, player)


@Packets.register(ServerPackets.CHANNEL_INFO)
def channel_info(stream: StreamIn, game: "Game"):
    name = stream.string()
    topic = stream.string()
    user_count = stream.s16()

    if not (c := game.bancho.channels.get(name)):
        game.bancho.channels.add(c := Channel(name, game, topic))

    c.user_count = user_count
    c.topic = topic

    if c.name == "#osu" and not c.joined:
        c.join()

    game.events.call(ServerPackets.CHANNEL_INFO, c)


@Packets.register(ServerPackets.CHANNEL_AUTO_JOIN)
def channel_autojoin(stream: StreamIn, game: "Game"):
    name = stream.string()
    topic = stream.string()
    user_count = stream.s16()

    if not (c := game.bancho.channels.get(name)):
        game.bancho.channels.add(c := Channel(name, game, topic))

    c.user_count = user_count
    c.topic = topic
    c.join_success()

    game.events.call(ServerPackets.CHANNEL_AUTO_JOIN, c)


@Packets.register(ServerPackets.CHANNEL_INFO_END)
def channel_info_end(stream: StreamIn, game: "Game"):
    game.events.call(ServerPackets.CHANNEL_INFO_END)


@Packets.register(ServerPackets.CHANNEL_JOIN_SUCCESS)
def channel_join_success(stream: StreamIn, game: "Game"):
    name = stream.string()

    if not (c := game.bancho.channels.get(name)):
        game.bancho.channels.add(c := Channel(name=name, game=game))

    c.join_success()

    game.events.call(ServerPackets.CHANNEL_JOIN_SUCCESS, c)


@Packets.register(ServerPackets.CHANNEL_KICK)
def channel_revoked(stream: StreamIn, game: "Game"):
    name = stream.string()

    if not (c := game.bancho.channels.get(name)):
        return

    game.bancho.channels.remove(c)
    game.logger.info(f"Kicked out of channel: {name}")
    game.events.call(ServerPackets.CHANNEL_KICK, c)


@Packets.register(ServerPackets.BEATMAP_INFO_REPLY)
def beatmapinfo_reply(stream: StreamIn, game: "Game"):
    beatmaps = [BeatmapInfo.decode(stream) for beatmap in range(stream.s32())]
    game.events.call(ServerPackets.BEATMAP_INFO_REPLY, beatmaps)


@Packets.register(ServerPackets.SILENCE_END)
def silence_info(stream: StreamIn, game: "Game"):
    if (remaining_silence := stream.s32()) > 0:
        game.bancho.player.silenced = True
        game.bancho.silenced = True
        game.logger.warning(f"You have been silenced for {remaining_silence} seconds.")

        threading.Timer(remaining_silence, game.bancho.unsilence)
    else:
        game.bancho.unsilence()

    game.events.call(ServerPackets.SILENCE_END, remaining_silence)


@Packets.register(ServerPackets.USER_SILENCED)
def user_silenced(stream: StreamIn, game: "Game"):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])

    if not player.loaded:
        game.bancho.request_presence([user_id])

    player.silenced = True

    game.logger.info(f"{player} has been silenced.")
    game.events.call(ServerPackets.USER_SILENCED, player)


@Packets.register(ServerPackets.TARGET_IS_SILENCED)
def target_silenced(stream: StreamIn, game: "Game"):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])

    if not player.loaded:
        game.bancho.request_presence([user_id])

    player.silenced = True

    game.logger.info(
        f"{player.name} has been silenced and is unable to respond to your messages right now."
    )
    game.events.call(ServerPackets.TARGET_IS_SILENCED, player)


@Packets.register(ServerPackets.USER_DM_BLOCKED)
def dms_blocked(stream: StreamIn, game: "Game"):
    user_id = stream.s32()

    if not (player := game.bancho.players.by_id(user_id)):
        game.bancho.request_presence([user_id])
        return

    player.dms_blocked = True

    game.logger.info(f"{player} blocked their dms.")
    game.events.call(ServerPackets.USER_DM_BLOCKED, player)



from concurrent.futures import ThreadPoolExecutor
from typing import List, Callable
from dataclasses import dataclass
from datetime import datetime
import logging


@dataclass
class Task:
    function: Callable
    interval: float
    loop: bool
    last_call: datetime
    threaded: bool


class TaskManager:
    """### TaskManager

    Used to run any task syncronously or asyncronously inside a thread.\n
    Example of registering a task that runs every minute:
    >>> @game.tasks.register(minutes=1, loop=True)
    >>> def example_task():
    >>>     ...
    """

    def __init__(self, game) -> None:
        self.executor = ThreadPoolExecutor(max_workers=10)
        self.tasks: List[Task] = []
        self.game = game

        self.logger = logging.getLogger("tasks")
        self.logger.disabled = game.logger.disabled

    def register(self, *, seconds=0, minutes=0, hours=0, loop=False, threaded=False):
        """Register a task

        `seconds`, `minutes`, `hours`: Specify when this task should run

        `loop`: If set to `True`, the task will loop itself in the specified time interval

        `threaded`: This will run the task inside a thread.
        """

        def wrapper(f: Callable):
            total = (seconds) + (minutes * 60) + (hours * 60 * 60)
            self.tasks.append(
                Task(
                    function=f,
                    interval=total,
                    loop=loop,
                    last_call=datetime.now(),
                    threaded=threaded,
                )
            )
            return f

        return wrapper

    def execute(self) -> None:
        """Execute every currently registered task"""
        if not self.game.bancho.connected:
            return

        for task in self.tasks:
            last_call = (datetime.now() - task.last_call).total_seconds()

            if last_call < task.interval:
                continue

            task.last_call = datetime.now()

            if not task.loop:
                self.tasks.remove(task)

            try:
                self.logger.debug(f"Trying to run task: '{task.function.__name__}'")

                if not task.threaded:
                    task.function()
                    self.logger.debug(
                        f"Task '{task.function.__name__}' was successfully executed."
                    )
                    continue

                if self.executor._shutdown:
                    return

                self.executor.submit(task.function)
                self.logger.debug(
                    f"Task '{task.function.__name__}' was submitted to executor."
                )

            except Exception as exc:
                self.logger.error(
                    f"Failed to run '{task.function.__name__}' task: {exc}",
                    exc_info=exc,
                )



from typing import Callable, Optional, Iterator, List, TYPE_CHECKING

from .constants import Mode, Mods, RankingType, CommentTarget, DisplayMode, ModeSelect
from ..objects.beatmap import OnlineBeatmap
from ..objects.score import ScoreResponse
from ..objects.comment import Comment

if TYPE_CHECKING:
    from ..game import Game

import requests
import logging


class WebAPI:
    """WebAPI
    -----------
    This class provides access to the client api endpoints, which includes:
        - Leaderboards
        - Comments
        - Friends
        - Replays
        - Avatars
        - Beatmap Star Ratings
        - Seasonal Backgrounds

    with the exception being score submission for obvious reasons.

    Functions:
        - `check_updates`
        - `connect`
        - `verify`
        - `get_session`
        - `get_backgrounds`
        - `get_friends`
        - `get_scores`
        - `get_star_rating`
        - `get_favourites`
        - `add_favourite`
        - `get_comments`
        - `post_comment`
        - `get_replay`
        - `get_avatar`
        - `get_beatmap_thumbnail`
        - `get_beatmap_preview`
        - `search_beatmapsets`
        - `download_osz`
    """

    def __init__(self, game: "Game") -> None:
        self.game = game

        self.session = requests.Session()
        self.session.headers = {"User-Agent": "osu!", "osu-version": self.game.version}

        self.logger = logging.getLogger(f"osu!api-{game.version}")
        self.logger.disabled = game.logger.disabled

        self.url = f"https://osu.{self.game.server}"
        self.asset_url = f"https://assets.{self.game.server}"

    @property
    def verification_url(self) -> str:
        return (
            f"{self.url}/p/verify"
            f'?u={self.game.username.replace(" ", "%20")}'
            f"&reason=bancho"
            f"&ch={self.game.client.hash}"
        )

    def connected_to_bancho(self, *args, **kwargs) -> Callable:
        """Can be used as a class wrapper to ensure that the client is connected"""

        def wrapper(f: Callable):
            if not self.game.bancho.connected:
                return

            return f(*args, **kwargs)

        return wrapper

    def check_updates(self) -> Optional[List[dict]]:
        """This will a request on `/web/check-updates.php`

        Returns:
            Optional[list]: When the request was successful, it will return a list of dicts with the following attributes:
                            file_version, filename, file_hash, filesize, timestamp, patch_id, url_full
        """
        self.logger.info("Checking for updates...")

        response = self.session.get(
            "https://osu.ppy.sh/web/check-updates.php",
            params={
                "action": "check",
                "stream": self.game.stream.lower(),
                "time": self.game.time,
            },
        )

        if not response.ok:
            self.logger.error(f"Failed to get updates ({response.status_code})")
            return None

        if "fallback" in response.text:
            self.logger.error(f'Failed to get updates: "{response.text}"')
            return None

        return response.json()

    def connect(self, retry=False) -> bool:
        """This will perform a request on `/web/bancho_connect.php`.

        Returns:
            bool: If the request was successful
        """

        self.logger.info("Connecting to bancho...")

        response = self.session.get(
            f"{self.url}/web/bancho_connect.php",
            params={
                "v": self.game.version,
                "u": self.game.username,
                "h": self.game.password_hash,
                "fx": "fail",  # dotnet version
                "ch": str(self.game.client.hash),
                "retry": int(retry),
            },
        )

        if not response.ok:
            if self.game.server == "ppy.sh":
                self.logger.error(
                    f"Error on login: {requests.status_codes._codes[response.status_code][0].upper()}"
                )
                self.logger.warning("Ignoring...")
            return True

        if "error" in response.text:
            self.logger.error(
                f'Error on login: {response.text.removeprefix("error: ")}'
            )

            if "verify" in response.text:
                self.verify()
                return False

        return True

    def verify(self, exit_after: bool = True) -> None:
        """This will print out a url, where the user can verify this client."""
        self.logger.info("Verification required.")
        self.logger.info(self.verification_url)
        self.logger.info("You only need to do this once.")

        if exit_after:
            exit(0)

    def get_session(self) -> requests.Response:
        """Perform a request on `/web/osu-session.php`.\n
        I don't know what this actually does.
        My guess is that it checks, if somebody is already online with this account.
        """  # TODO

        response = self.session.post(
            f"{self.url}/web/osu-session.php",
            files={
                "u": self.game.username,
                "h": self.game.password_hash,
                "action": "check",
            },
        )

        return response

    def get_backgrounds(self) -> Optional[list]:
        """This will perform a request on `/web/osu-getseasonal.php`.

        Returns:
            Optional[list]: List of urls for the seasonal background images
        """

        response = self.session.get(f"{self.url}/web/osu-getseasonal.php")

        if response.ok:
            return response.json()

    def get_menu_content(self) -> Optional[dict]:
        """This will perform a request on `/web/osu-getcurrent.php`.

        Returns:
            Optional[dict]: The menu content
        """

        response = self.session.get(f"{self.asset_url}/menu-content.json")

        if response.ok:
            return response.json()

    def get_friends(self) -> List[int]:
        """This will perform a request on `/web/osu-getfriends.php`.

        Returns:
            List[int]: The user ids of your friends
        """

        response = self.session.get(
            f"{self.url}/web/osu-getfriends.php",
            params={"u": self.game.username, "h": self.game.password_hash},
        )

        if response.ok:
            return [int(id) for id in response.text.split("\n") if id.isdigit()]

        return []

    def get_scores(
        self,
        beatmap_checksum: str,
        beatmap_file: str,
        set_id: int,
        mode: Mode = Mode.Osu,
        mods: Optional[Mods] = Mods.NoMod,
        rank_type=RankingType.Top,
        skip_scores: bool = False,
    ) -> Optional[ScoreResponse]:
        """Get top scores for a beatmap

        Args:
            beatmap_checksum (str): MD5 hash of the beatmap file
            beatmap_file (str): Filename of the beatmap
            set_id (int): BeatmapsetId for the beatmap
            mode (Mode, optional): Specify a mode
            mods (Mods, optional): Filter by mods (`rank_type` must be set to `SelectedMod`)
            rank_type (RankingType, optional): Select the leaderboard type

        Returns:
            Optional[ScoreResponse]
        """

        response = self.session.get(
            f"{self.url}/web/osu-osz2-getscores.php",
            params={
                "s": int(skip_scores),
                "vv": 4,  # ?
                "v": rank_type.value,
                "c": beatmap_checksum,
                "f": beatmap_file,
                "m": mode.value,
                "i": set_id,
                "mods": mods.value,
                "a": 0,  # ?
                "us": self.game.username,
                "ha": self.game.password_hash,
            },
        )

        if not response.ok:
            self.logger.error(f"Failed to fetch scores ({response.status_code})")
            return

        return ScoreResponse.from_string(response.text, mode)

    def get_star_rating(
        self, beatmap_id: int, mode: Mode = Mode.Osu, mods: Mods = Mods.NoMod
    ) -> float:
        """Get star rating of a beatmap

        Args:
            beatmap_id (int): Beatmap ID
            mode (Mode, optional): Mode
            mods (Mods, optional): Mods

        Returns: float
        """

        response = self.session.post(
            f"{self.url}/difficulty-rating",
            json={
                "beatmap_id": beatmap_id,
                "ruleset_id": mode.value,
                "mods": [{"acronym": acronym} for acronym in mods.acronyms],
            },
        )

        if not response.ok:
            return 0.0

        return float(response.text)

    def get_favourites(self) -> List[int]:
        """Get your beatmap favourites

        Returns:
            List[int]: List of beatmapset ids
        """

        response = self.session.get(
            f"{self.url}/web/osu-getfavourites.php",
            params={"u": self.game.username, "h": self.game.password_hash},
        )

        if not response.ok:
            return []

        return [
            int(beatmap_id) for beatmap_id in response.text.split("\n") if beatmap_id
        ]

    def add_favourite(self, beatmapset_id: int) -> str:
        """Add a beatmap to your favourites list

        Args:
            beatmapset_id (int)

        Returns:
            str: A human-readable response string. e.g. "Added to favourites! You have a total of 2 favourites."
        """

        response = self.session.get(
            f"{self.url}/web/osu-addfavourite.php",
            params={
                "u": self.game.username,
                "h": self.game.password_hash,
                "a": beatmapset_id,
            },
        )

        return response.text

    def get_comments(
        self,
        beatmap_id: Optional[int] = None,
        set_id: Optional[int] = None,
        replay_id: Optional[int] = None,
        mode: Mode = Mode.Osu,
    ) -> List[Comment]:
        """Get comments for a beatmap, set or replay

        Args:
            beatmap_id (int, optional): Get comments from this map
            set_id (int, optional): Get comments from this set
            replay_id (int, optional): Get comments from this replay
            mode (Mode, optional): Specify the mode

        Returns:
            List[Comment]
        """

        if all([beatmap_id is None, set_id is None, replay_id is None]):
            return []

        response = self.session.post(
            f"{self.url}/web/osu-comment.php",
            files={
                "u": (None, self.game.username),
                "p": (None, self.game.password_hash),
                "b": (None, beatmap_id),
                "s": (None, set_id),
                "r": (None, replay_id),
                "m": (None, mode.value),
                "a": (None, "get"),
            },
        )

        if not response.ok:
            self.logger.error(f"Failed to fetch comments ({response.status_code})")
            return []

        if not response.text:
            return []

        return [Comment.from_string(line) for line in response.text.split("\n") if line]

    def post_comment(
        self,
        text: str,
        time: int,
        target: CommentTarget = CommentTarget.Map,
        beatmap_id: Optional[int] = None,
        replay_id: Optional[int] = None,
        set_id: Optional[int] = None,
        mode: Mode = Mode.Osu,
    ) -> None:
        """Post a comment to a map, song or replay

        Args:
            text (str): The content of your comment
            time (int): The time you want this comment to appear inside the map
            target (CommentTarget): Specify where this comment should appear.
            mode (Mode, optional): Specify the mode
            beatmap_id (Optional[int], optional): Set the beatmap id, if you set the target to "Map"
            replay_id (Optional[int], optional): Set the replay id, if you set the target to "Replay"
            set_id (Optional[int], optional): Set the set it, if you set the target to "Song"
        """

        if all([beatmap_id is None, set_id is None, replay_id is None]):
            return

        target_id = {
            CommentTarget.Map: beatmap_id,
            CommentTarget.Song: set_id,
            CommentTarget.Replay: replay_id,
        }[target]

        if target_id is None:
            return

        self.session.post(
            f"{self.url}/web/osu-comment.php",
            files={
                "u": (None, self.game.username),
                "p": (None, self.game.password_hash),
                "b": (None, beatmap_id),
                "s": (None, set_id),
                "r": (None, replay_id),
                "m": (None, mode.value),
                "a": (None, "post"),
                "starttime": (None, time),
                "comment": (None, text),
                "target": (None, target.value),
            },
        )

    def get_replay(self, replay_id: int, mode: Mode = Mode.Osu) -> Optional[bytes]:
        """Get raw replay data by id (not osr!)

        Args:
            replay_id (int): The score/replay id
            mode (Mode, optional): The mode of this score

        Returns:
            Optional[bytes]: The raw replay data, compressed with lzma

        Please view this page for more information: https://osu.ppy.sh/wiki/de/Client/File_formats/osr_%28file_format%29
        """

        response = self.session.get(
            f"{self.url}/web/osu-getreplay.php",
            params={
                "u": self.game.username,
                "h": self.game.password_hash,
                "m": mode.value,
                "c": replay_id,
            },
        )

        if not response.ok:
            self.logger.error(f"Failed to fetch replay ({response.status_code})")
            return

        return response.content

    def get_avatar(self, user_id: int) -> Optional[bytes]:
        """Get avatar by user id"""

        return self.session.get(f"https://a.{self.game.server}/{user_id}").content

    def get_beatmap_thumbnail(
        self, beatmapset_id: int, large: bool = False
    ) -> Optional[bytes]:
        """Get the background thumbnail of a beatmap"""

        response = self.session.get(
            f"https://b.{self.game.server}/thumb/{beatmapset_id}{'l' if large else ''}.jpg"
        )

        return response.content if response.ok else None

    def get_beatmap_preview(self, beatmapset_id: int) -> Optional[bytes]:
        """Get the preview to a song of a beatmap"""

        response = self.session.get(
            f"https://b.{self.game.server}/preview/{beatmapset_id}.mp3"
        )

        return response.content if response.ok else None

    def download_osz(
        self, beatmapset_id: int, no_video: bool = False
    ) -> Optional[Iterator[bytes]]:
        """Download an osz file

        Args:
            beatmapset_id (int): The beatmapset id
            no_video (bool, optional): Specify if the osz should contain a video

        Returns:
            Optional[Iterator[bytes]]: An iterator of bytes, which contains the osz
        """

        response = self.session.get(
            f"https://osu.ppy.sh/d/{beatmapset_id}{'n' if no_video else ''}",
            allow_redirects=True,
            stream=True,
            params={
                "u": self.game.username,
                "h": self.game.password_hash,
                "vv": 2,  # what is this lol
            },
        )

        if not response.ok:
            return

        return response.iter_content(1024)

    def search_beatmapsets(
        self, query: str, display_mode=DisplayMode.Ranked, mode=ModeSelect.All, page=0
    ) -> Optional[List[OnlineBeatmap]]:
        """Get a list of beatmapsets, aka. osu! direct search

        Args:
            query (str): Search for beatmaps
            display_mode (DisplayMode): Filter maps by their status
            mode (ModeSelect): Filter maps by their mode
            page (int): Specify an offset/page

        Returns:
            Optional[List[OnlineBeatmap]]
        """

        response = self.session.get(
            f"https://osu.ppy.sh/web/osu-search.php",
            params={
                "u": self.game.username,
                "h": self.game.password_hash,
                "q": query,
                "r": display_mode.value,
                "m": mode.value,
                "p": page,
            },
        )

        lines = response.text.splitlines()
        status = int(lines[0])

        if status < 0:
            self.logger.error(f'Failed to get beatmapsets: "{lines[1]}"')
            return

        return [OnlineBeatmap.parse(line) for line in lines[1:]]



from ..bancho.constants import Grade, Mode
from ..bancho.streams import StreamIn

from typing import List, Optional, Tuple
from dataclasses import dataclass
from datetime import datetime


@dataclass
class BeatmapInfo:
    id: int
    beatmap_id: int
    beatmapset_id: int
    thread_id: int
    ranked: int
    osu_rank: Grade
    fruits_rank: Grade
    taiko_rank: Grade
    mania_rank: Grade
    checksum: str

    @classmethod
    def decode(cls, stream: StreamIn) -> "BeatmapInfo":
        return BeatmapInfo(
            stream.s16(),
            stream.s32(),
            stream.s32(),
            stream.s32(),
            stream.u8(),
            Grade(stream.u8()),
            Grade(stream.u8()),
            Grade(stream.u8()),
            Grade(stream.u8()),
            stream.string(),
        )


@dataclass
class OnlineBeatmap:
    osz_filename: str
    artist: str
    title: str
    creator: str
    status: int
    rating: float
    last_update: datetime
    set_id: int
    thread_id: int
    has_video: bool
    has_storyboard: bool
    filesize: int
    filesize_novideo: Optional[int]
    difficulties: List[Tuple[str, Mode]]

    @classmethod
    def parse(cls, string: str) -> "OnlineBeatmap":
        args = string.split("|")
        return OnlineBeatmap(
            args[0],
            args[1],
            args[2],
            args[3],
            int(args[4]),
            float(args[5]),
            datetime.fromisoformat(args[6]),
            int(args[7]),
            int(args[8]),
            args[9] == "1",
            args[10] == "1",
            int(args[11]),
            int(args[12]) if args[12] else None,
            [
                (difficulty.split("@")[0], Mode(int(difficulty.split("@")[-1])))
                for difficulty in args[13].split(",")
            ],
        )



import functools
import warnings


def deprecated(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        warnings.warn(
            f"{func.__name__} is deprecated and will be removed in a future version.",
            DeprecationWarning,
            stacklevel=2,
        )
        return func(*args, **kwargs)

    return wrapper



from typing import Optional, List, TYPE_CHECKING
from datetime import datetime
from queue import Queue

from ..bancho.client import BanchoClient as HTTPBanchoClient
from ..bancho.constants import Privileges, ServerPackets
from ..bancho.streams import StreamIn, StreamOut
from ..bancho.constants import ClientPackets
from ..bancho.packets import Packets

from ..objects.collections import Players, Channels
from ..objects.player import Player

if TYPE_CHECKING:
    from .game import TcpGame

import logging
import socket
import gzip
import time


class TcpBanchoClient(HTTPBanchoClient):
    def __init__(self, game: "TcpGame", ip: str, port: int) -> None:
        self.game = game

        self.logger = logging.getLogger(f"tcp-bancho-{game.version}")
        self.logger.disabled = game.logger.disabled

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.port = port
        self.ip = ip

        self.user_id = -1
        self.connected = False
        self.retry = True

        self.spectating: Optional[Player] = None
        self.player: Optional[Player] = None

        self.channels = Channels()
        self.players = Players(game)
        self.queue = Queue()

        self.ping_count = 0
        self.protocol = 0

        self.privileges: Privileges = Privileges.Normal
        self.friends: List[int] = []

        self.last_action = datetime.now().timestamp()
        self.silenced = False
        self.in_lobby = False

    def run(self):
        """Run the client loop"""
        self.connect()

        while self.connected:
            try:
                self.dequeue()
                self.game.tasks.execute()
            except KeyboardInterrupt:
                raise
            except Exception as exc:
                self.logger.fatal(f"Unhandled Exception: {exc}", exc_info=exc)

        if self.retry:
            self.logger.error("Retrying in 15 seconds...")
            time.sleep(15)
            self.game.run(retry=True)

    def connect(self):
        """Perform the initial connection to the server"""
        login_data = f"{self.game.username}\r\n{self.game.password_hash}\r\n{self.game.client}\r\n"

        try:
            self.socket.connect((self.ip, self.port))
            self.socket.send(login_data.encode())
            self.connected = True
        except ConnectionRefusedError:
            self.logger.error("Connection refused by the server.")

    def process_packets(self) -> None:
        """Process incoming packets from the server"""
        packet_header = StreamIn(self.socket.recv(7))

        if packet_header.eof():
            return

        packet_id = packet_header.u16()
        compression = packet_header.bool()
        packet_size = packet_header.u32()
        packet_data = self.socket.recv(packet_size)

        if compression:
            packet_data = gzip.decompress(packet_data)

        packet, stream = ServerPackets(packet_id), StreamIn(packet_data)

        self.logger.debug(f'Received packet {packet.name} -> "{stream.get()}"')
        self.game.packets.packet_received(packet, stream, self.game)

    def enqueue(
        self, packet: ClientPackets, data: bytes = b"", dequeue: bool = False
    ) -> bytes:
        """Send a packet to the server and dequeue"""
        stream = StreamOut()

        # Construct header
        stream.u16(packet.value)
        stream.bool(False)
        stream.u32(len(data))
        stream.write(data)

        self.logger.debug(f'Sending {packet.name} -> "{data}"')
        self.socket.send(stream.get())
        self.last_action = datetime.now().timestamp()

        if dequeue:
            self.dequeue()

        return stream.get()

    def dequeue(self) -> None:
        """Process all incoming packets from the server"""
        if not self.connected:
            return

        try:
            self.process_packets()
        except OverflowError as e:
            self.logger.error(f'Failed to process packets: "{e}"')
            self.connected = False
            self.retry = True

    def exit(self):
        """Send logout packet to bancho, and disconnect."""
        try:
            self.retry = False
            self.connected = False
            self.enqueue(ClientPackets.LOGOUT, int(0).to_bytes(4, "little"))
        except BrokenPipeError:
            pass



from typing import Optional, List, TYPE_CHECKING
from datetime import datetime
from queue import Queue

from .constants import ClientPackets, ReplayAction, StatusAction, Privileges
from .streams import StreamOut

from ..objects.replays import ReplayFrame, ScoreFrame
from ..objects.collections import Players, Channels
from ..objects.player import Player
from ..objects.status import Status

if TYPE_CHECKING:
    from ..game import Game

import requests
import logging
import time


class BanchoClient:
    """BanchoClient
    ---------------
    To interact with bancho, you can set event listeners on server packets.

    Here is an example for the message packet:
    ```
    @game.events.register(ServerPackets.SEND_MESSAGE)
    def on_message(sender: Player, message: str, target: Channel|Player):
        print(f"Received message from {sender.name} on {target.name}: {message}")
    ```
    Please refer to `osu.bancho.packets.py`, to find the arguments that are being passed into the event.

    You can also create tasks, that run in specific intervals:
    ```
    @game.tasks.register(minutes=1, loop=True)
    def example_task():
        # This code runs every minute
        ...
    ```

    Attributes:
        `player`: osu.objects.Player

        `spectating`: osu.objects.Player

        `players`: osu.objects.Players

        `channels`: osu.objects.Channels

        `privileges`: osu.bancho.constants.Privileges

        `friends`: list[int]

        `game`: osu.game.Game

        `connected`: bool

        `retry`: bool

        `min_idletime`: int (Minimum time between requests)

        `max_idletime`: int (Maximum time between requests)

    Functions:
        `enqueue`: Send a bancho packet to the server

        `ping`: Send a ping packet

        `request_presence`: Request a presence update for a list of players

        `request_stats`: Request a stats update for a list of players

        `request_status`: Request a status update for your account

        `update_status`: Update `player.status`

        `start_spectating`: Start spectating a player

        `stop_spectating`: Stop spectating

        `cant_spectate`: Is sent when the player does not have the beatmap for spectating

        `send_frames`: Send replay frames to your spectators

        `join_lobby`: Join the lobby

        `leave_lobby`: Leave the lobby
    """

    def __init__(self, game: "Game") -> None:
        self.game = game

        self.logger = logging.getLogger(f"bancho-{game.version}")
        self.logger.disabled = game.logger.disabled

        self.domain = f"c.{game.server}"
        self.url = f"https://{self.domain}"

        self.session = requests.Session()
        self.session.headers = {
            "osu-version": self.game.version,
            "Accept-Encoding": "gzip, deflate",
            "User-Agent": "osu!",
            "Host": self.domain,
        }

        self.user_id = -1
        self.connected = False
        self.retry = True
        self.token = ""

        self.spectating: Optional[Player] = None
        self.player: Optional[Player] = None

        self.channels = Channels()
        self.players = Players(game)
        self.queue = Queue()

        self.ping_count = 0
        self.protocol = 0

        self.privileges: Privileges = Privileges.Normal
        self.friends: List[int] = []

        self.last_action = datetime.now().timestamp()
        self.fast_read = False
        self.silenced = False
        self.in_lobby = False

        self.min_idletime = 1
        self.max_idletime = 2.5

    @property
    def status(self) -> Status:
        """Status of the connected player"""
        if not self.player:
            return Status()
        return self.player.status

    @property
    def idle_time(self) -> float:
        """Time between now and the last request"""
        return datetime.now().timestamp() - self.last_action

    @property
    def request_interval(self) -> int:
        """Time between requests"""
        if self.fast_read:
            return 0

        interval = 1

        if not self.spectating:
            interval *= 1 + self.idle_time / 10
            interval *= self.ping_count

        return min(self.max_idletime, max(self.min_idletime, interval))

    def run(self) -> None:
        """Run the client loop"""
        self.connect()

        while self.connected:
            try:
                time.sleep(self.request_interval)
                self.dequeue()
                self.game.tasks.execute()
            except KeyboardInterrupt:
                raise
            except Exception as exc:
                self.logger.fatal(f"Unhandled Exception: {exc}", exc_info=exc)

        if self.retry:
            self.logger.error("Retrying in 15 seconds...")
            time.sleep(15)
            self.game.run(retry=True)

    def connect(self) -> None:
        """Perform the initial connection to the bancho server, to get a connection token"""
        data = f"{self.game.username}\n{self.game.password_hash}\n{self.game.client}\n"

        response = self.session.post(self.url, data=data)

        if not response.ok:
            self.connected = False
            self.retry = True
            self.logger.error(
                f"[{response.url}]: Connection was refused ({response.status_code})"
            )
            return

        if not (token := response.headers.get("cho-token")):
            # Failed to get token
            self.connected = False
            self.retry = False
            self.game.packets.data_received(response.content, self.game)
            return

        self.connected = True
        self.token = token

        self.session.headers["osu-token"] = self.token
        self.game.packets.data_received(response.content, self.game)

    def dequeue(self) -> None:
        """Send a request to bancho, empty the queue and handle incoming packets"""

        if not self.connected:
            return

        if self.queue.empty():
            # Queue is empty, sending ping
            self.ping_count += 1
            return self.ping()

        if self.queue.qsize() > 1:
            self.ping_count = 0

        data = b""

        while not self.queue.empty():
            data += self.queue.get()

        response = self.session.post(self.url, data=data)

        if not response.ok:
            self.connected = False
            self.retry = True
            self.logger.error(
                f"[{response.request.url}]: Connection was refused ({response.status_code})"
            )
            return

        self.fast_read = False
        self.game.packets.data_received(response.content, self.game)
        self.last_action = datetime.now().timestamp()

    def exit(self) -> None:
        """Send logout packet to bancho, and disconnect."""
        self.enqueue(ClientPackets.LOGOUT, int(0).to_bytes(4, "little"))
        self.connected = False
        self.retry = False

    def enqueue(self, packet: ClientPackets, data: bytes = b"", dequeue=True) -> bytes:
        """Send a packet to the queue and dequeue"""

        stream = StreamOut()

        # Construct header
        stream.u16(packet.value)
        stream.bool(False)
        stream.u32(len(data))
        stream.write(data)

        self.logger.debug(f'Sending {packet.name} -> "{data}"')

        # Append to queue
        self.queue.put(stream.get())

        if dequeue:
            self.dequeue()

        return stream.get()

    def unsilence(self) -> None:
        """Will be called when the connected player gets unsilenced by the server"""
        if not self.silenced:
            return

        self.player.silenced = False
        self.silenced = False
        self.logger.info("You can now talk again.")

    def ping(self) -> None:
        """Send a ping to the server"""
        self.enqueue(ClientPackets.PING)

    def request_presence(self, ids: List[int]) -> None:
        """Request the presence of a list of players"""
        stream = StreamOut()
        stream.intlist(ids)

        self.enqueue(ClientPackets.USER_PRESENCE_REQUEST, stream.get())

    def request_stats(self, ids: List[int]) -> None:
        """Request the stats of a list of players"""
        stream = StreamOut()
        stream.intlist(ids)

        self.enqueue(ClientPackets.USER_STATS_REQUEST, stream.get())

    def request_status(self) -> None:
        """Request a status update for the connected player"""
        self.enqueue(ClientPackets.REQUEST_STATUS_UPDATE)

    def update_status(self) -> None:
        """Send current player status to the server

        You can change your status by updating the `game.bancho.player.status` attributes.\n
        For example:
            >>> status = game.bancho.player.status
            >>>
            >>> status.action = StatusAction.Playing
            >>> status.text = "UNDEAD CORPORATION - Everything will freeze [Time Freeze]"
            >>> status.checksum = "a84050da9b68ca1bd8e2d1700b9c6ca5" # .osu file checksum
            >>> status.mods = Mods.Hidden|Mods.HardRock
            >>> status.mode = GameMode.Osu
            >>> status.beatmap_id = 555797
            >>>
            >>> game.bancho.update_status()
        """

        if not self.player:
            return

        stream = StreamOut()
        stream.u8(self.player.status.action.value)
        stream.string(self.player.status.text)
        stream.string(self.player.status.checksum)
        stream.u32(self.player.status.mods.value)
        stream.u8(self.player.mode.value)
        stream.s32(self.player.status.beatmap_id)

        self.enqueue(ClientPackets.CHANGE_ACTION, stream.get())

    def start_spectating(self, target: Player) -> None:
        """Start spectating a player

        You will receive `ServerPackets.SPECTATE_FRAMES` packets that contain replay data.\n
        You can use osu-recorder (https://github.com/Lekuruu/osu-recorder) as a reference for that.
        """
        if self.spectating:
            self.stop_spectating()

        self.enqueue(
            ClientPackets.START_SPECTATING,
            int(target.id).to_bytes(4, "little"),
            dequeue=False,
        )

        self.spectating = target

        target.request_presence()
        target.request_stats()

        self.status.action = StatusAction.Watching
        self.status.text = target.status.text
        self.status.checksum = target.status.checksum
        self.status.mods = target.status.mods
        self.status.mode = target.status.mode
        self.status.beatmap_id = target.status.beatmap_id

        self.update_status()

    def stop_spectating(self) -> None:
        """Stop spectating a player"""
        if not self.spectating:
            return

        self.logger.info(f"Stopped spectating {self.spectating.name}.")

        self.enqueue(ClientPackets.STOP_SPECTATING, dequeue=False)
        self.spectating = None

        self.status.reset()
        self.update_status()

    def cant_spectate(self) -> None:
        """Sends a packet to let other spectators know that you have a missing beatmap"""
        self.enqueue(ClientPackets.CANT_SPECTATE)

    def send_frames(
        self,
        action: ReplayAction,
        frames: List[ReplayFrame],
        score_frame: Optional[ScoreFrame] = None,
        seed: int = 0,
    ) -> None:
        """
        This will send a `SPECTATE_FRAMES` packet to the server, which gets broadcasted to all of your spectators.
        Note that this will not work, if nobody is spectating you.
        """
        if not self.player.spectators:
            self.logger.warning(
                "Failed to send frames, because no spectators were found."
            )
            return

        if self.spectating:
            action = ReplayAction.WatchingOther
            extra = self.spectating.id
        else:
            extra = seed

        stream = StreamOut()
        stream.s32(extra)
        stream.u16(len(frames))
        [stream.write(frame.encode()) for frame in frames]
        stream.u8(action.value)

        if score_frame:
            stream.write(score_frame.encode())

        self.enqueue(ClientPackets.SPECTATE_FRAMES, stream.get())

    def join_lobby(self) -> None:
        """Join the multiplayer lobby"""
        if self.in_lobby:
            return

        self.enqueue(ClientPackets.JOIN_LOBBY)
        self.in_lobby = True

    def leave_lobby(self) -> None:
        """Leave the multiplayer lobby"""
        if not self.in_lobby:
            return

        self.enqueue(ClientPackets.PART_LOBBY)
        self.in_lobby = True
